/*
* Copyright (c) 2005-2006
* Authors:
*   Godefroid Chapelle <gotcha@bubblenet.be>
*   Florian Schulze <florian.schulze@gmx.net>
*   Balázs Reé <ree@greenfinity.hu>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as published
* by the Free Software Foundation.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
* 02111-1307, USA.
*/

/* Event registration */

kukit.er = {};

/* Event registry 
*
*  available for plugin registration
*
*  usage:
*
*  kukit.er.eventRegistry.register(namespace, eventname, func, 
*    bindmethodname, defaultactionmethodname);
*  
*     namespace = null: means global namespace
*     defaultactionmethodname = null: if there is no default action implemented
*     func must be a class (constructor) function, this is the class that
*           implements the binder.
*
*/

kukit.er.eventClassCounter = 0;

kukit.er.EventRegistry = function () {
    this.content = {};
    this.classes = {};
};

/* binder registration */

kukit.er.EventRegistry.prototype.registerBinder = function(classname, func) {
    if (typeof(func) == 'undefined') {
        throw 'Func is mandatory.';
    }
    if (this.classes[classname]) {
        // Do not allow redefinition
        kukit.logError('Error : redefinition attempt of event class ' + classname);
        return;
    
    }
    // Decorate and store the class
    kukit.er.decorateEventBinderClass(func);
    this.classes[classname] = func;
};

kukit.er.EventRegistry.prototype.existsBinder = function(classname) {
    var func = this.classes[classname];
    return (typeof(func) != 'undefined');
};

kukit.er.EventRegistry.prototype.getBinder = function(classname) {
    var func = this.classes[classname];
    if (! func) {
        // not found
        throw 'Error : undefined event setup type ' + classname;
        }
    return func;
};

/* events (methods) registration */

kukit.er.EventRegistry.prototype.register = function(namespace, eventname, klass,
        bindmethodname, defaultactionmethodname) {
    if (typeof(defaultactionmethodname) == 'undefined') {
        throw 'EventRegistry.register misses some parameters';
    }
    // Find out the class name. (Not specified now.)
    var classname = klass.prototype.__classname__;
    if (typeof(classname) == 'undefined') {
        // Create a classname, and register it too.
        classname = '' + kukit.er.eventClassCounter;
        kukit.er.eventClassCounter += 1;
        this.registerBinder(classname, klass);
        klass.prototype.__classname__ = classname;
    }
    if (!eventname) {
        throw 'In EventRegistry.register eventname must be non-empty';
    }
    var key = this._getKey(namespace, eventname);
    var entry = this.content[key];
    if (typeof(entry) != 'undefined') {
        if (key[0] == '-') {
            key = key.substring(1);
        }
        throw 'In EventRegistry.register double registration of key "' + key + '"';
    }
    // register it
    this.content[key] = {
        'classname': classname,
        'bindmethodname': bindmethodname,
        'defaultactionmethodname': defaultactionmethodname
        };
};

kukit.er.EventRegistry.prototype._getKey = function(namespace, eventname) {
    if (namespace == null) {
        namespace = '';
    } else if (namespace.split('-') > 1) {
        throw 'In EventRegistry.register namespace cannot contain -';
    }
    return namespace + '-' + eventname;
}

kukit.er.EventRegistry.prototype.exists = function(namespace, eventname) {
    var key = this._getKey(namespace, eventname);
    var entry = this.content[key];
    return (typeof(entry) != 'undefined');
};

kukit.er.EventRegistry.prototype.get = function(namespace, eventname) {
    var key = this._getKey(namespace, eventname);
    var entry = this.content[key];
    if (typeof(entry) == 'undefined') {
        if (key[0] == '-') {
            key = key.substring(1);
            throw 'Error : undefined global event key ' + key + ' (or maybe namespace is missing?)';
        } else {
            throw 'Error : undefined event key ' + key;
        }
    } 
    return entry;
};

kukit.er.eventRegistry = new kukit.er.EventRegistry();

/* Event class decoration 
*
* poor man's subclassing
* This is called automatically on registration, to dress
* up the event class with the necessary methods
*
*/

kukit.er.EventBinder_calldefault = function(namespace, name, oper) {
    // private method for calling just a default method.
    // params should be alredy prepared.
    var methodname = kukit.er.eventRegistry.get(namespace, name).defaultactionmethodname;
    if (methodname) {
        var method = this[methodname];
        if (! method) {
            throw 'Could not bind event name "' + name + '" on namespace "' + namespace + '", because the method "' + methodname + '" does not exist.';
            }
        // call it
        oper.binderinstance = this;
        method.call(this, name, oper);
    }
}

/* This can be called on the event and will execute the
*  desired event action. Parameters will be these + those
*  defined in the rule.
*
* Call example: binderinstance.__trigger_event__('doit', {'extravalue': '5'}, node);
*/

kukit.er.EventBinder__trigger_event__ = function(name, parms, node) {
    // Called programmatically
    //
    // XXX uhh this must go away. An implicit kss rule is only to be executed
    // if no other nodes execute.
    var executed = 0;
    var special = 0;
    // document rules. Execute them with node = null
    eventrule =  kukit.rd.methodTable.getMergedRule('document', name, this);
    if (eventrule != null) {
        var oper = new kukit.op.Oper();
        oper.parms = parms;
        oper.eventrule = eventrule;
        // node is null.
        this._EventBinder_triggerevent(name, oper);
        special += 1;
    }
    // Normal rules. If any of those match, execute them too
    // each on the node that it selects - not on the original node.
    var opers = kukit.er.binderInstanceRegistry.getBinding(this, name);
    for (var i=0; i<opers.length; i++) {
        var oper = opers[i];
        var newoper = new kukit.op.Oper();
        newoper.parms = parms;
        newoper.node = oper.node;
        newoper.eventrule = oper.eventrule;
        this._EventBinder_triggerevent(name, newoper);
        executed += 1;
    }
    // look up the behaviour event rule. This may fail (eventrule = null)
    // but yet we want to execute a default action in this case.
    var eventrule =  kukit.rd.methodTable.getMergedRule('behaviour', name, this);
    // XXX uhh this must go away. If we have no eventrule, we must remember
    // it and execute but only if otherwise not bound (this is an implicit
    if (eventrule != null || executed + special == 0) {
        // behaviour method is executed on the old node
        // XXX bad, very bad
        var oper = new kukit.op.Oper();
        oper.parms = parms;
        oper.node = node;
        oper.eventrule = eventrule;
        this._EventBinder_triggerevent(name, oper);
        special += 1;
    }
    kukit.logDebug('Event "' + name + '" executed on ' + executed + ' nodes, ' + special + ' special kss targets.');
}

kukit.er.EventBinder_triggerevent = function(name, oper) {
    // Private. Called from __trigger_event__ or from main event execution.
    if (oper.eventrule) {
        // Call the actions, if we had an event rule.
        // This includes calling the default action.
        oper.binderinstance = this;
        oper.eventrule.actions.execute(oper);
    } else {
        // In case there is no event rule, just call the default event action.
        // This is programmatic calling.
        var namespace = this.__event_namespace__;
        kukit.logDebug('Calling implicit event "' + name + '" on namespace "' + namespace + '"')
        this._EventBinder_calldefault(namespace, name, oper);
    }
};

kukit.er.decorateEventBinderClass = function(cls) {
    cls.prototype.__trigger_event__ = kukit.er.EventBinder__trigger_event__;
    cls.prototype._EventBinder_triggerevent = kukit.er.EventBinder_triggerevent;
    cls.prototype._EventBinder_calldefault = kukit.er.EventBinder_calldefault;
};

/* Event instance registry 
*
*  used in run-time to keep track of the event instances
*
*/

kukit.er.BinderInstanceRegistry = function () {
    this.content = {};
};

kukit.er.BinderInstanceRegistry.prototype.getOrCreateBinder = function (id, classname, namespace) {
    // Get or create the event.
    var binderinstance = this.content[id];
    if (typeof(binderinstance) == 'undefined') {
        // Create a new event.
        kukit.logDebug('instantiating event id=' + id + ', classname=' + classname + ', namespace=' + namespace);
        binderinstance = this.content[id] = new (kukit.er.eventRegistry.getBinder(classname))();
        // decorate it with id and class
        binderinstance.__binder_id__ = id;
        binderinstance.__binder_classname__ = classname;
        binderinstance.__event_namespace__ = namespace;
        binderinstance.__event_rules__ = [];
    } else if (binderinstance.__binder_classname__ != classname) {
        // just paranoia
        throw 'Conflicting class for event id "' + id + '", "' + binderinstance.__binder_classname__
                + '" != "' + classname + '"';
    }
    return binderinstance;
};

kukit.er.BinderInstanceRegistry.prototype.markBinding = function (oper) {
    // Marks that the instance was bound to this rule
    var name = oper.eventrule.kss_selector.name;
    rules_already_bound = oper.binderinstance.__event_rules__[name];
    if (typeof(rules_already_bound) == 'undefined') {
        // Create an empty list.
        rules_already_bound = oper.binderinstance.__event_rules__[name] = [];
    }
    rules_already_bound.push(oper);
};

kukit.er.BinderInstanceRegistry.prototype.getBinding = function (binderinstance, name) {
    // Get the eventrule, node items that bind to the instance with an event name
    var opers = binderinstance.__event_rules__[name];
    if (typeof(opers) == 'undefined') {
        // Create an empty list.
        opers = binderinstance.__event_rules__[name] = [];
    }
    // Return it
    return opers;
};

kukit.er.BinderInstanceRegistry.prototype.getBinderById = function (id) {
    // Get an event.
    var binderinstance = this.content[id];
    if (typeof(binderinstance) == 'undefined') {
        throw 'Event with id "' + id + '" not found.'
    }
    return binderinstance;
};

kukit.er.BinderInstanceRegistry.prototype.getSingletonBinderByName = function (namespace, name) {
    //Get classname
    var classname = kukit.er.eventRegistry.get(namespace, name).classname;
    // Get an event.
    var id = kukit.rd.makeId(namespace, classname);
    var binderinstance = this.content[id];
    if (typeof(binderinstance) == 'undefined') {
        throw 'Singleton event with namespace "' + namespace + '" and (event) name "' + name + '" not found.'
    }
    return binderinstance;
};

kukit.er.binderInstanceRegistry = new kukit.er.BinderInstanceRegistry();
